from collections.abc import Iterator
from typing import Any

from django.contrib import admin
from django.contrib.admin.options import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.core.validators import EMPTY_VALUES
from django.db.models import Max, Min, Model, QuerySet
from django.db.models.fields import (
    AutoField,
    DecimalField,
    Field,
    FloatField,
    IntegerField,
)
from django.forms import ValidationError
from django.http import HttpRequest

from unfold.contrib.filters.admin.mixins import RangeNumericMixin
from unfold.contrib.filters.forms import SingleNumericForm, SliderNumericForm


class SingleNumericFilter(admin.FieldListFilter):
    request = None
    parameter_name = None
    template = "unfold/filters/filters_numeric_single.html"

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)

        if not isinstance(field, DecimalField | IntegerField | FloatField | AutoField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request

        if self.parameter_name is None:
            self.parameter_name = self.field_path

        if self.parameter_name in params:
            value = params.pop(self.parameter_name)
            value = value[0] if isinstance(value, list) else value

            if value not in EMPTY_VALUES:
                self.used_parameters[self.parameter_name] = value

    def queryset(
        self, request: HttpRequest, queryset: QuerySet[Any]
    ) -> QuerySet | None:
        if self.value() and self.parameter_name:
            try:
                return queryset.filter(**{self.parameter_name: self.value()})
            except (ValueError, ValidationError):
                return None

    def value(self) -> Any:
        if self.parameter_name:
            return self.used_parameters.get(self.parameter_name)

    def expected_parameters(self) -> list[str | None]:
        return [self.parameter_name]

    def choices(self, changelist: ChangeList) -> Iterator:
        if self.parameter_name:
            yield {
                "request": self.request,
                "parameter_name": self.parameter_name,
                "form": SingleNumericForm(
                    name=self.parameter_name, data={self.parameter_name: self.value()}
                ),
            }


class RangeNumericListFilter(RangeNumericMixin, admin.SimpleListFilter):
    def __init__(
        self,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
    ) -> None:
        super().__init__(request, params, model, model_admin)

        self.request = request
        self.init_used_parameters(params)

    def lookups(
        self, request: HttpRequest, model_admin: ModelAdmin
    ) -> tuple[tuple[str, str], ...]:
        return (("dummy", "dummy"),)


class RangeNumericFilter(RangeNumericMixin, admin.FieldListFilter):
    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)
        if not isinstance(field, DecimalField | IntegerField | FloatField | AutoField):
            raise TypeError(
                f"Class {type(self.field)} is not supported for {self.__class__.__name__}."
            )

        self.request = request
        if self.parameter_name is None:
            self.parameter_name = self.field_path

        self.init_used_parameters(params)


class SliderNumericFilter(RangeNumericFilter):
    MAX_DECIMALS = 7
    STEP = None

    template = "unfold/filters/filters_numeric_slider.html"
    field = None
    form_class = SliderNumericForm

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        super().__init__(field, request, params, model, model_admin, field_path)

        self.field = field
        self.q = model_admin.get_queryset(request)

    def choices(self, changelist: ChangeList) -> Iterator:
        total = self.q.all().count()
        min_value = self.q.all().aggregate(min=Min(self.parameter_name)).get("min", 0)
        max_value = None

        if total > 1:
            max_value = (
                self.q.all().aggregate(max=Max(self.parameter_name)).get("max", 0)
            )

        decimals = 0
        step = self.STEP if self.STEP else 1

        if isinstance(self.field, FloatField | DecimalField):
            decimals = self.MAX_DECIMALS
            step = self.STEP if self.STEP else self._get_min_step(self.MAX_DECIMALS)

        if self.parameter_name:
            yield {
                "decimals": decimals,
                "step": step,
                "parameter_name": self.parameter_name,
                "request": self.request,
                "min": min_value,
                "max": max_value,
                "value_from": self.used_parameters.get(
                    self.parameter_name + "_from", min_value
                ),
                "value_to": self.used_parameters.get(
                    self.parameter_name + "_to", max_value
                ),
                "form": self.form_class(
                    name=self.parameter_name,
                    min=min_value,
                    max=max_value,
                    data={
                        self.parameter_name + "_from": self.used_parameters.get(
                            self.parameter_name + "_from", min_value
                        ),
                        self.parameter_name + "_to": self.used_parameters.get(
                            self.parameter_name + "_to", max_value
                        ),
                    },
                ),
            }

    def _get_min_step(self, precision: int) -> float:
        result_format = f"{{:.{precision - 1}f}}"
        return float(result_format.format(0) + "1")



from typing import Any

from django.core.validators import EMPTY_VALUES
from django.forms import MultiWidget, Widget
from django.http import QueryDict
from django.utils.datastructures import MultiValueDict

from unfold.widgets import (
    PROSE_CLASSES,
    UnfoldAdminSelectWidget,
    UnfoldAdminTextInputWidget,
)

WYSIWYG_CLASSES = [
    *PROSE_CLASSES,
    "border!",
    "border-base-200!",
    "border-t-0!",
    "group-[.errors]:border-red-600",
    "max-w-none",
    "p-4",
    "rounded-b",
    "rounded-t-none",
    "text-base-500",
    "w-full",
    "focus:outline-hidden",
    "dark:border-base-700!",
    "dark:text-base-300!",
    "dark:group-[.errors]:border-red-500!",
]


class ArrayWidget(MultiWidget):
    template_name = "unfold/forms/array.html"

    def __init__(
        self,
        widget_class: type[Widget] | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.choices = kwargs.get("choices")
        self.widget_class = widget_class

        attrs = kwargs.get("attrs", {})
        attrs.update(
            {
                "class": attrs.get("class", "") if attrs else "",
            }
        )

        widgets = [self.get_widget_instance()]
        super().__init__(widgets, attrs=attrs)

    def get_widget_instance(self) -> Any:
        if hasattr(self, "widget_class") and self.widget_class is not None:
            return self.widget_class()

        if hasattr(self, "choices") and self.choices is not None:
            return UnfoldAdminSelectWidget(choices=self.choices)

        return UnfoldAdminTextInputWidget()

    def get_context(self, name: str, value: Any, attrs: dict | None) -> dict:
        self._resolve_widgets(value)
        context = super().get_context(name, value, attrs)
        context.update(
            {"template": self.get_widget_instance().get_context(name, "", {})["widget"]}
        )
        return context

    def value_from_datadict(
        self, data: QueryDict, files: MultiValueDict, name: str
    ) -> list:  # ty:ignore[invalid-method-override]
        values = []

        for item in data.getlist(name):
            if item not in EMPTY_VALUES:
                values.append(item)

        return values

    def value_omitted_from_data(
        self, data: QueryDict, files: MultiValueDict, name: str
    ) -> bool:  # ty:ignore[invalid-method-override]
        return data.getlist(name) not in [[""], *EMPTY_VALUES]

    def decompress(self, value: Any) -> list:
        if isinstance(value, list):
            return value
        elif isinstance(value, str):
            return value.split(",")

        return []

    def _resolve_widgets(self, value: list | str | None) -> None:
        if value is None:
            value = []
        elif isinstance(value, list):
            self.widgets = [self.get_widget_instance() for item in value]
        else:
            self.widgets = [self.get_widget_instance() for item in value.split(",")]

        self.widgets_names = ["" for i in range(len(self.widgets))]
        self.widgets = [w if isinstance(w, type) else w for w in self.widgets]


class WysiwygWidget(Widget):
    template_name = "unfold/forms/wysiwyg.html"

    class Media:
        css = {"all": ("unfold/forms/css/trix/trix.css",)}
        js = (
            "unfold/forms/js/trix/trix.js",
            "unfold/forms/js/trix.config.js",
        )

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        super().__init__(attrs)

        self.attrs.update(
            {
                "class": " ".join(
                    [*WYSIWYG_CLASSES, attrs.get("class", "") if attrs else ""]
                )
            }
        )



from collections.abc import Iterator

from django.contrib import admin
from django.contrib.admin.options import ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.db.models import Field, Model
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from unfold.contrib.filters.admin.mixins import ValueMixin
from unfold.contrib.filters.forms import SearchForm


class TextFilter(admin.SimpleListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = SearchForm

    def has_output(self) -> bool:
        return True

    def lookups(self, request: HttpRequest, model_admin: ModelAdmin) -> tuple:
        return ()

    def choices(self, changelist: ChangeList) -> Iterator:
        yield {
            "form": self.form_class(
                name=self.parameter_name,
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                data={self.parameter_name: self.value()},
            ),
        }


class FieldTextFilter(ValueMixin, admin.FieldListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = SearchForm

    def __init__(
        self,
        field: Field,
        request: HttpRequest,
        params: dict[str, str],
        model: type[Model],
        model_admin: ModelAdmin,
        field_path: str,
    ) -> None:
        self.lookup_kwarg = f"{field_path}__icontains"
        self.lookup_val = params.get(self.lookup_kwarg)
        super().__init__(field, request, params, model, model_admin, field_path)

    def expected_parameters(self) -> list[str | None]:
        return [self.lookup_kwarg]

    def choices(self, changelist: ChangeList) -> Iterator:
        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                data={self.lookup_kwarg: self.value()},
            ),
        }



from typing import Any

from django.http import HttpRequest


class ComponentRegistry:
    _registry: dict[str, type] = {}

    @classmethod
    def register_class(cls, component_cls: type) -> None:
        if not issubclass(component_cls, BaseComponent):
            raise ValueError(
                f"Class '{component_cls.__name__}' must inherit from BaseComponent."
            )

        class_name = component_cls.__name__

        if class_name in cls._registry:
            raise ValueError(f"Class '{class_name}' is already registered.")

        cls._registry[class_name] = component_cls

    @classmethod
    def get_class(cls, class_name: str) -> type | None:
        return cls._registry.get(class_name)

    @classmethod
    def create_instance(cls, class_name: str, **kwargs: Any) -> Any:
        component_cls = cls.get_class(class_name)

        if component_cls is None:
            raise ValueError(f"Class '{class_name}' is not registered.")

        return component_cls(**kwargs)


def register_component(cls: type) -> type:
    ComponentRegistry.register_class(cls)
    return cls


class BaseComponent:
    def __init__(self, request: HttpRequest):
        self.request = request

    def get_context_data(self, **kwargs):
        return kwargs



from unfold.mixins.action_model_admin import ActionModelAdminMixin
from unfold.mixins.base_model_admin import BaseModelAdminMixin
from unfold.mixins.dataset_model_admin import DatasetModelAdminMixin

__all__ = ["BaseModelAdminMixin", "ActionModelAdminMixin", "DatasetModelAdminMixin"]



from typing import Any

import django
from django.contrib import messages
from django.contrib.admin.views.main import ERROR_FLAG, PAGE_VAR
from django.contrib.admin.views.main import ChangeList as BaseChangeList
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import HttpRequest, JsonResponse
from django.views.generic import ListView

from unfold.exceptions import UnfoldException
from unfold.forms import DatasetChangeListSearchForm


class ChangeList(BaseChangeList):
    def __init__(self, request: HttpRequest, *args: Any, **kwargs: Any) -> None:
        super().__init__(request, *args, **kwargs)

        if django.VERSION < (5, 0):
            self.filter_params = dict(request.GET.lists())
            self.filter_params.pop(PAGE_VAR, None)
            self.filter_params.pop(ERROR_FLAG, None)


class DatasetChangeList(ChangeList):
    is_dataset = True

    def __init__(self, request: HttpRequest, *args: Any, **kwargs: Any) -> None:
        self.search_var = f"{kwargs.get('model')._meta.model_name}-q"
        self.page_var = f"{kwargs.get('model')._meta.model_name}-p"

        _search_form = DatasetChangeListSearchForm(
            request.GET, search_var=self.search_var
        )
        if not _search_form.is_valid():
            for error in _search_form.errors.values():
                messages.error(request, ", ".join(error))

        self.dataset_search_query = _search_form.cleaned_data.get(self.search_var) or ""

        super().__init__(request, *args, **kwargs)

    def get_results(self, request: HttpRequest) -> None:
        try:
            self.page_num = int(request.GET.get(self.page_var, 1))
        except ValueError:
            self.page_num = 1

        super().get_results(request)

    def get_queryset(self, request, exclude_parameters=None):
        self.query = self.dataset_search_query
        return super().get_queryset(request, exclude_parameters)


class UnfoldModelAdminViewMixin(PermissionRequiredMixin):
    """
    Prepares views to be displayed in admin
    """

    model_admin = None

    def __init__(self, model_admin, **kwargs):
        self.model_admin = model_admin
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        self.request.current_app = self.model_admin.admin_site.name

        if not hasattr(self, "model_admin"):
            raise UnfoldException(
                "UnfoldModelAdminViewMixin was not provided with 'model_admin' argument"
            )

        if not hasattr(self, "title"):
            raise UnfoldException(
                "UnfoldModelAdminViewMixin was not provided with 'title' attribute"
            )

        return super().get_context_data(
            **kwargs,
            **self.model_admin.admin_site.each_context(self.request),
            **{
                "title": self.title,
                "model_admin": self.model_admin,
            },
        )


class BaseAutocompleteView(ListView):
    paginate_by = 20

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> JsonResponse:
        super().get(request, *args, **kwargs)
        context = self.get_context_data()

        return JsonResponse(
            {
                "results": [
                    {
                        "id": obj.pk,
                        "text": str(obj),
                    }
                    for obj in context["object_list"]
                ],
                "pagination": {
                    "more": context["page_obj"].has_next(),
                },
            }
        )



from typing import Any

from django.contrib.admin.checks import ModelAdminChecks
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.auth.models import Permission
from django.core import checks

from unfold.dataclasses import UnfoldAction


class UnfoldModelAdminChecks(ModelAdminChecks):
    def check(self, admin_obj: BaseModelAdmin, **kwargs) -> list[checks.Error]:
        return [
            *super().check(admin_obj, **kwargs),
            *self._check_unfold_action_permission_methods(admin_obj),
        ]

    def _check_unfold_action_permission_methods(self, obj: Any) -> list[checks.Error]:
        """
        Actions with an allowed_permission attribute require the ModelAdmin to
        implement a has_<perm>_permission() method for each permission.
        """
        actions: list[UnfoldAction] = [
            *obj._get_base_actions_list(),
            *obj._get_base_actions_detail(),
            *obj._get_base_actions_row(),
            *obj._get_base_actions_submit_line(),
        ]
        errors = []
        for action in actions:
            if not hasattr(action.method, "allowed_permissions"):
                continue

            for permission in action.method.allowed_permissions:
                # Check the existence of Django permission
                if "." in permission:
                    app_label, codename = permission.split(".")

                    if not Permission.objects.filter(
                        content_type__app_label=app_label,
                        codename=codename,
                    ).exists():
                        errors.append(
                            checks.Error(
                                f"@action decorator on {action.method.original_function_name}() in class {obj.__class__.__name__} specifies permission {permission} which does not exists.",
                                obj=obj.__class__,
                                id="admin.E129",
                            )
                        )

                    continue

                # Check the permission method existence
                method_name = f"has_{permission}_permission"
                if not hasattr(obj, method_name):
                    errors.append(
                        checks.Error(
                            f"{obj.__class__.__name__} must define a {method_name}() method for the {action.method.original_function_name}() action.",
                            obj=obj.__class__,
                            id="admin.E129",
                        )
                    )

        return errors



from collections.abc import Iterator
from typing import Any

from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.db.models import Model
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from unfold.admin import ModelAdmin
from unfold.contrib.filters.admin.mixins import (
    ChoicesMixin,
    MultiValueMixin,
    ValueMixin,
)
from unfold.contrib.filters.forms import CheckboxForm, HorizontalRadioForm, RadioForm


class RadioFilter(admin.SimpleListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = RadioForm
    all_option = ["", _("All")]

    def choices(self, changelist: ChangeList) -> Iterator:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        choices = []

        if self.all_option:
            choices = [self.all_option]

        if add_facets:
            for i, (lookup, title) in enumerate(self.lookup_choices):
                choices.append(
                    (lookup, f"{title} ({facet_counts.get(f'{i}__c', '-')})")
                )
        else:
            choices.extend(self.lookup_choices)

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.parameter_name,
                choices=choices,
                data={self.parameter_name: self.value()},
            ),
        }


class CheckboxFilter(RadioFilter):
    form_class = CheckboxForm
    all_option = None

    # TODO: remove once django 4.x is not supported
    def __init__(
        self,
        request: HttpRequest,
        params: dict[str, Any],
        model: type[Model],
        model_admin: ModelAdmin,
    ) -> None:
        self.request = request
        super().__init__(request, params, model, model_admin)

    def value(self) -> list[str] | None:  # ty:ignore[invalid-method-override]
        if self.parameter_name:
            return self.request.GET.getlist(self.parameter_name)


class ChoicesRadioFilter(ValueMixin, ChoicesMixin, admin.ChoicesFieldListFilter):
    form_class = RadioForm
    all_option = ["", _("All")]


class ChoicesCheckboxFilter(
    MultiValueMixin, ChoicesMixin, admin.ChoicesFieldListFilter
):
    form_class = CheckboxForm
    all_option = None


class BooleanRadioFilter(ValueMixin, admin.BooleanFieldListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = HorizontalRadioForm
    all_option = ["", _("All")]

    def choices(self, changelist: ChangeList) -> Iterator:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets and facet_counts:
            choices = [
                self.all_option,
                *[
                    ("1", f"{_('Yes')} ({facet_counts['true__c']})"),
                    ("0", f"{_('No')} ({facet_counts['false__c']})"),
                ],
            ]
        else:
            choices = [
                self.all_option,
                *[
                    ("1", _("Yes")),
                    ("0", _("No")),
                ],
            ]

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }


class RelatedCheckboxFilter(MultiValueMixin, admin.RelatedFieldListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = CheckboxForm

    def choices(self, changelist: ChangeList) -> Iterator:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets and facet_counts:
            choices = []

            for pk_val, val in self.lookup_choices:
                count = facet_counts[f"{pk_val}__c"]
                choice = (pk_val, f"{val} ({count})")
                choices.append(choice)
        else:
            choices = self.lookup_choices

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }


class AllValuesCheckboxFilter(MultiValueMixin, admin.AllValuesFieldListFilter):
    template = "unfold/filters/filters_field.html"
    form_class = CheckboxForm

    def choices(self, changelist: ChangeList) -> Iterator:
        add_facets = getattr(changelist, "add_facets", False)
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None

        if add_facets and facet_counts:
            choices = []

            for i, val in enumerate(self.lookup_choices):
                count = facet_counts[f"{i}__c"]
                choice = (val, f"{val} ({count})")
                choices.append(choice)
        else:
            choices = [[val, val] for _i, val in enumerate(self.lookup_choices)]

        yield {
            "form": self.form_class(
                label=_(" By %(filter_title)s ") % {"filter_title": self.title},
                name=self.lookup_kwarg,
                choices=choices,
                data={self.lookup_kwarg: self.value()},
            ),
        }
